typedef struct csr_frag csr_frag;
typedef struct csr_target csr_target;

struct csr_frag
{
	u32 id;	// Triangle index
	float depth;	// 'depth testing'
	
	v3f co;
	v3f nrm;
};

struct csr_target
{
	csr_frag *fragments;
	u32 x, y; 
	v4f bounds;
};

void csr_rt_clear( csr_target *rt )
{
	for( u32 i = 0; i < rt->x*rt->y; i ++ )
	{
		rt->fragments[ i ].depth = 0.f;
	}
}

void simple_raster( csr_target *rt, vmf_vert tri[3], int id )
{
	// Very simplified tracing algorithm
	float tqa = 0.f, tqb = 0.f;
	
	v2f bmin = { 0.f, 0.f };
	v2f bmax = { rt->x, rt->y };
	
	v2_minv( tri[0].co, tri[1].co, bmin );
	v2_minv( tri[2].co, bmin, bmin );
	
	v2_maxv( tri[0].co, tri[1].co, bmax );
	v2_maxv( tri[2].co, bmax, bmax );

	float range_x = (rt->bounds[2]-rt->bounds[0])/(float)rt->x;
	float range_y = (rt->bounds[3]-rt->bounds[1])/(float)rt->y;
	
	int start_x = csr_min( rt->x-1, csr_max( 0,     floorf( (bmin[0]-rt->bounds[0])/range_x)));
	int end_x =   csr_max( 0,       csr_min( rt->x-1, floorf( (bmax[0]-rt->bounds[0])/range_x )));
	int start_y = csr_min( rt->y-1, csr_max( 0,      ceilf( (bmin[1]-rt->bounds[1])/range_y )));
	int end_y =   csr_max( 0,       csr_min( rt->y-1,  ceilf( (bmax[1]-rt->bounds[1])/range_y )));
	
	v3f trace_dir = { 0.f, 0.f, 1.f };
	v3f trace_origin = { 0.f, 0.f, -16385.f };

	for( u32 py = start_y; py <= end_y; py ++ )
	{
		trace_origin[1] = csr_lerpf( rt->bounds[1], rt->bounds[3], (float)py/(float)rt->y );
		
		for( u32 px = start_x; px <= end_x; px ++ )
		{
			csr_frag *frag = &rt->fragments[ py * rt->y + px ];

			trace_origin[0] = csr_lerpf( rt->bounds[0], rt->bounds[2], (float)px/(float)rt->x );
			float tdepth = csr_ray_tri( trace_origin, trace_dir, tri[0].co, tri[1].co, tri[2].co, &tqa, &tqb );
			
			if( tdepth > frag->depth )
			{
				frag->depth = tdepth;
				
				v3_muls( tri[1].co, tqa, frag->co );
				v3_muladds( frag->co, tri[2].co, tqb, frag->co );
				v3_muladds( frag->co, tri[0].co, 1.f - tqa - tqb, frag->co );
			}
		}
	}
}

void csr_draw( csr_target *rt, vmf_vert *triangles, u32 triangle_count, m4x3f transform )
{
	m3x3f normal;
	vmf_vert new_tri[3];

	// Derive normal matrix
	m4x3_to_3x3( transform, normal );
	m3x3_inv_transpose( normal, normal );

	for( u32 i = 0; i < triangle_count; i ++ )
	{
		vmf_vert *triangle = triangles + i*3;
		
		m4x3_mulv( transform, triangle[0].co, new_tri[0].co );
		m4x3_mulv( transform, triangle[1].co, new_tri[1].co );
		m4x3_mulv( transform, triangle[2].co, new_tri[2].co );
		m3x3_mulv( normal, triangle[0].nrm, new_tri[0].nrm );
		m3x3_mulv( normal, triangle[1].nrm, new_tri[1].nrm );
		m3x3_mulv( normal, triangle[2].nrm, new_tri[2].nrm );
		
		simple_raster( rt, new_tri, 0 );
	}
}

void draw_vmf_group( csr_target *rt, vmf_map *map, vdf_node *root, int const group, m4x3f prev, m4x3f inst )
{
	m4x3f transform = M4X3_IDENTITY;
	vmf_solid solid;
	vmf_vert tri[3];

	// Multiply previous transform with instance transform to create basis
	if( prev )
	{
		m4x3_mul( prev, inst, transform );
	}

	// Draw brushes
	solidgen_ctx_init( &solid );
	vdf_node *world = vdf_next( root, "world", NULL );
	
	vdf_foreach( world, "solid", brush )
	{
		solidgen_push( &solid, brush );
	}

	for( int i = 0; i < csr_sb_count( solid.indices )/3; i ++ )
	{
		u32 * base = solid.indices + i*3;
		
		tri[0] = solid.verts[ base[0] ];
		tri[1] = solid.verts[ base[1] ];
		tri[2] = solid.verts[ base[2] ];
		
		csr_draw( rt, tri, 1, transform );
	}
	
	solidgen_ctx_reset( &solid );
		
	// Actual entity loop
	m4x3f model;
	
	vdf_foreach( root, "entity", ent )
	{
		if( ent->user & VMF_FLAG_IS_PROP )
		{
			// Create model transform
			m4x3_identity( model );
			
			vmf_entity_transform( ent, model );
			m4x3_mul( transform, model, model );
			
			// Draw model
			mdl_mesh_t *mdl = &map->models[ ent->user1 ].mdl;
			for( int i = 0; i < mdl->num_indices/3; i ++ )
			{
				for( int j = 0; j < 3; j ++ )
				{
					v3_copy( &mdl->vertices[ mdl->indices[ i*3+j ] *8 ],   tri[j].co );
					v3_copy( &mdl->vertices[ mdl->indices[ i*3+j ] *8+3 ], tri[j].nrm );
					tri[j].xy[0] = 0.f;
					tri[j].xy[1] = 0.f;
				}
				
				csr_draw( rt, tri, 1, model );
			}
		} 
		else if( ent->user & VMF_FLAG_IS_INSTANCE )
		{
			m4x3_identity( model );
			vmf_entity_transform( ent, model );
			
			draw_vmf_group( rt, map, map->cache[ ent->user1 ].root, group, transform, model );
		}
		else if( ent->user & VMF_FLAG_BRUSH_ENT )
		{
			// ...
		}
	}
	
	solidgen_ctx_free( &solid );
}
