#define vdf_foreach( NODE, STR, AS ) \
int __vdf_it_##AS = 0; \
vdf_node * AS;\
while( (AS = vdf_next( NODE, STR, &__vdf_it_##AS )) )

#define kv_foreach( NODE, STR, AS ) \
int __kv_it_##AS = 0; \
const char * AS;\
while( (AS = kv_iter( NODE, STR, &__kv_it_##AS )) )

#include <stdio.h>
#include <stdint.h>
#include <string.h>
#include <ctype.h>

// TYPES
// ==================================================================================================================

typedef struct vdf_kv vdf_kv;
typedef struct vdf_node vdf_node;
typedef struct vdf_ctx vdf_ctx;

// API
// ==================================================================================================================

// Open / close
vdf_node *vdf_open_file( const char *fn );
void vdf_free_r( vdf_node *p );

// Nodes
// -----

// Starting from *it, get next child with matching name from node.
vdf_node *vdf_next( vdf_node *node, const char *name, int *it );

// Create new empty node attached to parent. name can be NULL
vdf_node *vdf_create_node( vdf_node *parent, const char *name );

// Save / printing
// ---------------

void vdf_out( vdf_node *h, int lvl, int declare, FILE *file );
void vdf_save( vdf_node *node, const char *fn );
void vdf_print( vdf_node *node );

// Keyvalue pairs
// --------------

// Get value string pointer from node's dictionary
const char *kv_get( vdf_node *node, const char *key, const char *value_defalt );

// Iterate each keyvalue starting from *it until key is matched
char *kv_iter( vdf_node *node, const char *key, int *it );

// Get keyvalue from node as int / float
int kv_get_int( vdf_node *node, const char *key, const int default_value );
float kv_get_float( vdf_node *node, const char *key, float default_value );

// Parse values into sepecifc type
void kv_int_array( vdf_node *node, const char *key, u32 count, int *arr );
void kv_float_array( vdf_node *node, const char *key, u32 count, float *arr );
void kv_double_array( vdf_node *node, const char *key, u32 count, double *arr );

// INTERNAL API
// ==================================================================================================================

// Add keyvalue pair to node
void vdf_kv_append( vdf_node *p, const char *k, const char *v );

// (low level api for arrays)
void kv_parse_array( const char *source, u32 esize, u32 count, void(*interp_func)(const char *src, void *dest), void *arr );
void vdf_str_to_float( const char *src, void *dest );
void vdf_str_to_int( const char *src, void *dest );

// Parsing context
// ---------------
void vdf_newln( vdf_ctx *ctx );
void vdf_endl( vdf_ctx *ctx );
int vdf_line_control( vdf_ctx *ctx );
void vdf_wait_endl( vdf_ctx *ctx );
void vdf_parse_string( vdf_ctx *ctx );
int vdf_parse_structure( vdf_ctx *ctx );
void vdf_parse_begin_token( vdf_ctx *ctx, char *ptr );
void vdf_parse_feedbuffer( vdf_ctx *ctx, char *buf );

// Formatting
void vdf_out_indent( const int n, FILE *file );

// IMPLEMENTATION
// ==================================================================================================================

struct vdf_kv
{
	char *key;
	char *value;
};

struct vdf_node
{
	char *name;
	
	vdf_node *parent;
	
	vdf_node **nodes;
	vdf_kv 	*pairs;
	
	u32		user;
	u32		user1;
};

vdf_node *vdf_next( vdf_node *node, const char *name, int *it )
{	
	if( !node ) 
		return NULL;

	for( int i = it? *it: 0; i < csr_sb_count( node->nodes ); i ++ )
	{
		if( !name || !strcmp( name, node->nodes[i]->name ))
		{
			if( it ) *it = i+1;
			return node->nodes[i];
		}
	}
	
	return NULL;
}

const char *kv_get( vdf_node *node, const char *key, const char *value_defalt )
{
	if( node ) 
	{
		for( int i = 0; i < csr_sb_count( node->pairs ); i ++ )
		{
			if( !strcmp( node->pairs[ i ].key, key ) )
			{
				return node->pairs[ i ].value;
			}
		}
	}
	
	return value_defalt;
}

char *kv_iter( vdf_node *node, const char *key, int *it )
{
	char *val;

	if( node )
	{
		while( *it < csr_sb_count( node->pairs ) )
		{
			if( !strcmp( node->pairs[ *it ].key, key ) )
			{
				val = node->pairs[ *it ].value;
				*it = *it + 1;
				return val;
			}
			
			*it = *it + 1;
		}
	}
	
	return NULL;
}

void vdf_str_to_int( const char *src, void *dest )
{
	*((int *)dest) = atoi( src );
}

void vdf_str_to_float( const char *src, void *dest )
{
	*((float *)dest) = atof( src );
}

void vdf_str_to_double( const char *src, void *dest )
{
	*((double *)dest) = atof( src );
}

void kv_parse_array( const char *source, u32 esize, u32 count, void(*interp_func)(const char *src, void *dest), void *arr )
{
	if( !source )
		return;

	char value_buf[ 64 ];
	int token = 0;
	
	u32 i = 0;
	u32 k = 0;
	
	char const *c = source;
	
	while( *c )
	{
		if( *c == ' ' || *c == '\t' || *c == '[' || *c == ']' || *c == '(' || *c == ')' )
		{
			if( token )
			{
				value_buf[ k ] = 0x00;
				token = 0;
				
				interp_func( value_buf, ((u8 *)arr) + i*esize );
				i ++;
				
				if( i >= count )
				{
					break;
				}
			}
		}
		else
		{
			if( !token )
			{
				token = 1;
				k = 0;
			}
			
			if( token )
			{
				if( k < sizeof( value_buf ) - 1 )
				{
					value_buf[ k ++ ] = *c;
				}
			}
		}
	
		c ++;
	}
	
	// Add remaining case if we hit null
	if( token && (i < count) )
	{
		value_buf[ k ] = 0x00;
		interp_func( value_buf, ((u8 *)arr) + i*esize );
	}
}

void kv_int_array( vdf_node *node, const char *key, u32 count, int *arr )
{
	kv_parse_array( kv_get( node, key, NULL ), sizeof(int), count, vdf_str_to_int, arr );
}

void kv_float_array( vdf_node *node, const char *key, u32 count, float *arr )
{
	kv_parse_array( kv_get( node, key, NULL ), sizeof(float), count, vdf_str_to_float, arr );
}

void kv_double_array( vdf_node *node, const char *key, u32 count, double *arr )
{
	kv_parse_array( kv_get( node, key, NULL ), sizeof(double), count, vdf_str_to_double, arr );
}

int kv_get_int( vdf_node *node, const char *key, const int default_value )
{
	const char *v = kv_get( node, key, NULL );
	return v? atoi(v): default_value;
}

float kv_get_float( vdf_node *node, const char *key, float default_value )
{
	const char *v = kv_get( node, key, NULL );
	return v? atof( v ): default_value;
}

vdf_node *vdf_create_node( vdf_node *parent, const char *name )
{
	vdf_node *node = calloc( 1, sizeof( vdf_node ) );
	
	if( name )
	{
		node->name = csr_malloc( strlen( name )+1 );
		strcpy( node->name, name );
	}
	
	if( parent )
	{
		node->parent = parent;
		
		parent->nodes = csr_sb_reserve( parent->nodes, 1, sizeof(vdf_node *) );
		
		vdf_node **child = (vdf_node **)csr_sb_use( parent->nodes );
		*child = node;
	}
	
	return node;
}

void vdf_kv_append( vdf_node *p, const char *k, const char *v )
{
	p->pairs = csr_sb_reserve( p->pairs, 1, sizeof(vdf_kv) );
	vdf_kv *kv = (vdf_kv *)csr_sb_use( p->pairs );

	u32 sv = strlen(v)+1;
	u32 sk = strlen(k)+1;
	
	kv->key = csr_malloc( sv+sk );
	kv->value = kv->key+sk;
	
	memcpy( kv->key, k, sk );
	memcpy( kv->value, v, sv );
}

void vdf_free_r( vdf_node *p )
{
	for( int i = 0; i < csr_sb_count( p->pairs ); i ++ )
	{
		free( p->pairs[ i ].key );
	}

	for( int i = 0; i < csr_sb_count( p->nodes ); i ++ )
	{
		vdf_free_r( p->nodes[ i ] );
	}
	
	csr_sb_free( p->pairs );
	csr_sb_free( p->nodes );
	free( p->name );
	free( p );
}

// PARSING
// ==================================================================================================================

struct vdf_ctx
{
	char name[2048];
	
	vdf_node *root;
	
	u32 lines;
	u32 errors;
	
	// State
	struct 
	{
		int wait;
		int expect_decl;

		char *tokens[2];
		int i;
		
		char *ptr_read;
		
		vdf_node *pnode;
	} 
	st;
};

void vdf_newln( vdf_ctx *ctx )
{
	ctx->lines ++;

	ctx->st.tokens[0] = NULL;
	ctx->st.tokens[1] = NULL;
	ctx->st.i = 0;
}

void vdf_endl( vdf_ctx *ctx )
{
	// Final out-tokens
	if( ctx->st.tokens[0] )
	{
		// Keypair
		if( ctx->st.tokens[1] )
		{
			vdf_kv_append( ctx->st.pnode, ctx->st.tokens[0], ctx->st.tokens[1] );
		}
		else
		{
			// Decl
			strcpy( ctx->name, ctx->st.tokens[0] );
			ctx->st.expect_decl = 1;
		}
	}
	
	vdf_newln( ctx );
}

int vdf_line_control( vdf_ctx *ctx )
{
	if( *ctx->st.ptr_read == '\r' )
	{
		*ctx->st.ptr_read = 0x00;
		return 1;
	}
	if( *ctx->st.ptr_read == '\n' )
	{
		*ctx->st.ptr_read = 0x00;
		vdf_endl( ctx );
		return 2;
	}
	
	return 0;
}

void vdf_wait_endl( vdf_ctx *ctx )
{
	while( (*ctx->st.ptr_read) && (*ctx->st.ptr_read != '\n') )
	{
		if( vdf_line_control( ctx ) == 2 )
		{
			return;
		}
	
		ctx->st.ptr_read ++;
	}
}

void vdf_parse_string( vdf_ctx *ctx )
{
	while( *ctx->st.ptr_read )
	{
		if( *ctx->st.ptr_read == '"' )
		{
			*ctx->st.ptr_read = 0x00;
			return;
		}
		
		if( vdf_line_control( ctx ) )
		{
			fprintf( stderr, "Unexpected end of line character (Line: %u)\n", ctx->lines );
			return;
		}
	
		ctx->st.ptr_read ++;
	}
}

int vdf_parse_structure( vdf_ctx *ctx )
{
	if( *ctx->st.ptr_read == '{' )
	{
		if( ctx->st.tokens[0] || !ctx->st.expect_decl )
		{
			fprintf( stderr, "Unexpected token '{' (Line: %u)\n", ctx->lines );
			ctx->errors ++;
		}
		
		ctx->st.expect_decl = 0;
		ctx->st.pnode = vdf_create_node( ctx->st.pnode, ctx->name );
		
		vdf_wait_endl( ctx );
		return 1;
	}
	
	// Closing block, jump read head back to parent
	if( *ctx->st.ptr_read == '}' )
	{
		if( !ctx->st.pnode->parent )
		{
			fprintf( stderr, "Unexpected token '}' (Line: %u)\n", ctx->lines );
			ctx->errors ++;
		}
		else
		{
			ctx->st.pnode = ctx->st.pnode->parent;
		}
		
		vdf_wait_endl( ctx );
		return 1;
	}
	
	return 0;
}

void vdf_parse_begin_token( vdf_ctx *ctx, char *ptr )
{
	ctx->st.tokens[ ctx->st.i ] = ptr;

	if( ctx->st.expect_decl )
	{
		fprintf( stderr, "Unexpected token '%s' (Line: %u)\n", ctx->name, ctx->lines );
		ctx->errors ++;
	}
}

void vdf_parse_feedbuffer( vdf_ctx *ctx, char *buf )
{
	ctx->st.ptr_read = buf;
	
	while( *ctx->st.ptr_read )
	{
		if( !vdf_line_control( ctx ) )
		{
			if( (*ctx->st.ptr_read == '/') && (ctx->st.ptr_read[1] == '/') )
			{
				*ctx->st.ptr_read = 0x00;
				ctx->st.ptr_read += 2;
				
				vdf_endl( ctx );
				vdf_wait_endl( ctx );
			}
			else
			{
				if( !vdf_parse_structure( ctx ) )
				{
					if( *ctx->st.ptr_read == ' ' || *ctx->st.ptr_read == '\t' )
					{
						*ctx->st.ptr_read = 0x00;

						if( ctx->st.tokens[ ctx->st.i ] )
						{
							ctx->st.i ++;
							
							if( ctx->st.i == 2 ) 
							{
								vdf_wait_endl( ctx );
							}
						}
					}
					// Start new entry
					else if( !ctx->st.tokens[ ctx->st.i ] )
					{
						if( *ctx->st.ptr_read == '"' )
						{
							*ctx->st.ptr_read = 0x00;
							ctx->st.ptr_read ++;							
							
							vdf_parse_begin_token( ctx, ctx->st.ptr_read );
							vdf_parse_string( ctx );
						}
						else
						{
							if( !( *ctx->st.ptr_read == '/' && *(ctx->st.ptr_read + 1) == *ctx->st.ptr_read ) )
							{	
								vdf_parse_begin_token( ctx, ctx->st.ptr_read );
							}
						}
					}
				}
			}
		}
		
		ctx->st.ptr_read ++;
	}
}

int vdf_load_into( const char *fn, vdf_node *node )
{
	char *text_src = csr_textasset_read( fn );
	
	if( !text_src )
	{
		fprintf( stderr, "vdf open failed\n" );
		return 0;
	}
	
	vdf_ctx ctx = {0};
	ctx.root = ctx.st.pnode = node;
	
	vdf_newln( &ctx );
	vdf_parse_feedbuffer( &ctx, text_src );
	free( text_src );
	
	return 1;
}

vdf_node *vdf_open_file( const char *fn )
{	
	vdf_node *root = vdf_create_node( NULL, NULL );
	if( vdf_load_into( fn, root ) )
	{
		return root;
	}
	else
	{
		vdf_free_r( root );
		return NULL;
	}
}

// OUTPUT
// ==================================================================================================================

void vdf_out_indent( const int n, FILE *file )
{
	for( int x = 0; x < n; x ++ ) 
		fprintf( file, "\t" );
}

void vdf_out( vdf_node *h, int lvl, int declare, FILE *file )
{
	if( declare )
	{ 
		vdf_out_indent( lvl, file ); fprintf( file, "\"%s\"\n", h->name );
		vdf_out_indent( lvl, file ); fprintf( file, "{\n" );
	}
	
	for( int i = 0; i < csr_sb_count( h->pairs ); i ++ )
	{
		vdf_out_indent( lvl+1, file ); fprintf( file, "\"%s\" \"%s\"\n", h->pairs[i].key, h->pairs[i].value );
	}
	
	for( int i = 0; i < csr_sb_count( h->nodes ); i ++ )
	{
		vdf_out( h->nodes[i], lvl + 1, 1, file );
	}
	
	if( declare )
	{
		vdf_out_indent( lvl, file ); fprintf( file, "}\n" );
	}
}

void vdf_save( vdf_node *node, const char *fn )
{
	FILE* file = fopen( fn, "w" );
	
	vdf_out( node, -1, 0, file );
	
	fclose( file );
}

void vdf_print( vdf_node *node )
{
	vdf_out( node, -1, 0, stdout );
}
