// Abstract Valve file system 
//=======================================================================================================================

struct valve_filesystem
{
	char gamedir[ 512 ];
	char exedir[ 512 ];
	char bindir[ 512 ];		// TODO: This is currently not set?
	
	VPKHeader *vpk;
	char **searchpaths;

	FILE *current_archive;
	u16 	current_idx;
} 
fs_global;

void fs_set_gameinfo( const char *path )
{
	struct valve_filesystem *fs = &fs_global;
	
	vdf_node *info = vdf_open_file( path );
	if( !info ) 
		return;	
	
	// Set gamedir
	strcpy( fs->gamedir, path );
	csr_path_winunix( fs->gamedir );
	*csr_findext( fs->gamedir, '/' ) = 0x00;
	
	// Set exe dir
	strcpy( fs->exedir, fs->gamedir );
	strcat( fs->exedir, "../" );
	
	// Get all search paths from file
	vdf_node *search_paths = vdf_next(vdf_next(vdf_next( info, "GameInfo", NULL ), "FileSystem", NULL ), "SearchPaths", NULL );
	
	kv_foreach( search_paths, "Game", kv_game )
	{
		if( kv_game[0] == '|' ) continue; //TODO: deal with engine replacements?? maybe??
	
		char *buf;
		if( csr_path_is_abs( kv_game ) )
		{
			buf = csr_malloc( strlen( kv_game ) + 2 );
			strcpy( buf, kv_game );
			strcat( buf, "/" );
		}
		else
		{
			buf = csr_malloc( strlen( fs->exedir ) + strlen( kv_game ) + 2 );
			strcpy( buf, fs->exedir );
			strcat( buf, kv_game );
			strcat( buf, "/" );
		}
		
		fs->searchpaths = csr_sb_reserve( fs->searchpaths, 1, sizeof( char * ) );
		fs->searchpaths[ csr_sb_count( fs->searchpaths ) ] = buf;
		csr_sb_use( fs->searchpaths );
	}
	
	vdf_free_r( info );
	
	// Look for pak01_dir
	char pack_path[512];
	for( int i = 0; i < csr_sb_count( fs->searchpaths ); i ++ )
	{
		strcpy( pack_path, fs->searchpaths[i] );
		strcat( pack_path, "pak01_dir.vpk" );
	
		if( (fs->vpk = (VPKHeader *)csr_asset_read( pack_path )) )
		{
			break;
		}
	}
	
	if( !fs->vpk )
	{
		fprintf( stderr, "Could not locate pak01_dir.vpk in %i searchpaths. Stock models will not load!", csr_sb_count( fs->searchpaths ) );
	}

	printf( "fs_info:\n\
  gamedir: %s\n\
   exedir: %s\n\
      bin: %s\n\
     pack: %s\n\
  searchpaths:\n", fs->gamedir, fs->exedir, fs->bindir, pack_path );
  
	for( int i = 0; i < csr_sb_count( fs->searchpaths ); i ++ )
	{
		printf( "    %s\n", fs->searchpaths[i] );  
	}
}

void fs_exit(void)
{
	struct valve_filesystem *fs = &fs_global;

	for( int i = 0; i < csr_sb_count( fs->searchpaths ); i ++ )
	{
		free( fs->searchpaths[ i ] );
	}
	csr_sb_free( fs->searchpaths );
	fs->searchpaths = NULL;

	vpk_free( fs->vpk );
	fs->vpk = NULL;
	
	if( fs->current_archive )
	{
		fclose( fs->current_archive );
		fs->current_archive = NULL;
	}
}

char *valve_fs_get( const char *path )
{
	struct valve_filesystem *fs = &fs_global;
	
	VPKDirectoryEntry *entry;
	char pak[ 533 ];

	if( (entry = vpk_find( fs->vpk, path )) )
	{
		if( entry->ArchiveIndex != fs->current_idx )
		{
			if( fs->current_archive )
			{
				fclose( fs->current_archive );
				fs->current_archive = NULL;
			}
			
			fs->current_idx = entry->ArchiveIndex;
		}
		
		if( !fs->current_archive )
		{
			sprintf( pak, "%scsgo/pak01_%03hu.vpk", fs->exedir, fs->current_idx );
			fs->current_archive = fopen( pak, "rb" );
			
			if( !fs->current_archive )
			{
				fprintf( stderr, "Could not locate %s\n", pak );
				return NULL;
			}
		}
		
		char *filebuf = csr_malloc( entry->EntryLength );
		
		fseek( fs->current_archive, entry->EntryOffset, SEEK_SET );
		if( fread( filebuf, 1, entry->EntryLength, fs->current_archive ) == entry->EntryLength )
		{
			return filebuf;
		}
		else
		{
			free( filebuf );
			return NULL;
		}
	}
	else
	{
		// Use physical searchpaths
		char path_buf[ 512 ];
		
		for( int i = 0; i < csr_sb_count( fs->searchpaths ); i ++ )
		{
			strcpy( path_buf, fs->searchpaths[ i ] );
			strcat( path_buf, path );
			
			char *filebuf;
			if( (filebuf = csr_asset_read( path_buf )) )
			{
				return filebuf;
			}
		}
		
		return NULL;
	}
}
